import linecache
import sys
import typing as t
from dataclasses import dataclass
from re import search
from types import FrameType
from warnings import warn

import dill
import msgspec
from aiopath import AsyncPath
from blake3 import blake3  # type: ignore
from itsdangerous import Serializer as SecureSerializer
from pydantic import SecretStr


@dataclass
class Serializers:
    json: t.Any = msgspec.json
    yaml: t.Any = msgspec.yaml
    msgpack: t.Any = msgspec.msgpack
    pickle: t.Any = dill
    toml: t.Any = msgspec.toml

    def __post_init__(self) -> None:
        self.pickle.encode = dill.dumps
        self.pickle.decode = dill.loads


serializers = Serializers()


class AcbEncoder:
    json: t.Callable[..., t.Any]
    yaml: t.Callable[..., t.Any]
    msgpack: t.Callable[..., t.Any]
    toml: t.Callable[..., t.Any]
    pickle: t.Callable[..., t.Any]
    serializers: dict[str, t.Callable[..., t.Any]]
    path: t.Optional[AsyncPath] = None
    action: t.Optional[str] = None
    sort_keys: bool = True
    use_list: bool = False
    secure: bool = False
    secret_key: str = ""
    secure_salt: str = ""
    serializer: t.Optional[t.Callable[..., t.Any]] = None

    def __init__(self) -> None:
        self.serializers = serializers.__dict__
        for s in self.serializers:
            setattr(self, s, self.__call__)

    async def process(
        self,
        obj: t.Any,
        **kwargs,
    ) -> bytes | int | None:
        if self.action in ("load", "decode"):
            if self.serializer is msgspec.msgpack:
                kwargs["use_list"] = self.use_list
            if isinstance(obj, AsyncPath):
                obj = await obj.read_text()
            return self.serializer.decode(obj, **kwargs)  # type: ignore
        elif self.action in ("dump", "encode"):
            # if self.serializer is msgspec.yaml:
            #     kwargs["sort_keys"] = self.sort_keys
            data = self.serializer.encode(obj, **kwargs)  # type: ignore
            if isinstance(self.path, AsyncPath):
                return await self.path.write_bytes(data)
            return data

    def get_vars(self, frame: FrameType) -> tuple[str, t.Any]:
        code_context = linecache.getline(frame.f_code.co_filename, frame.f_lineno)
        calling_method = search(r"await\s(\w+)\.(\w+)\(", code_context)
        return calling_method.group(1), calling_method.group(2)  # type: ignore

    def get_serializer(
        self,
        serializer: t.Any,
    ) -> t.Any:
        return (
            serializer
            if not self.secure
            else SecureSerializer(
                secret_key=self.secret_key,
                salt=self.secure_salt,
                serializer=serializer,
                signer_kwargs=dict(digest_method=blake3),
            )
        )

    async def __call__(
        self,
        obj: t.Any,
        path: t.Optional[AsyncPath] = None,
        sort_keys: bool = True,
        use_list: bool = False,
        secret_key: t.Optional[SecretStr] = None,
        secure_salt: t.Optional[SecretStr] = None,
        **kwargs,
    ) -> dict | bytes:
        self.path = path
        self.sort_keys = sort_keys
        self.use_list = use_list
        self.secret_key = "" if not secret_key else secret_key.get_secret_value()
        self.secure_salt = "" if not secure_salt else secure_salt.get_secret_value()
        self.action, serializer_name = self.get_vars(sys._getframe(1))
        self.secure = all((secret_key, secure_salt))
        if not self.secure and (len(self.secret_key) and len(self.secure_salt)):
            warn(
                f"{serializer_name.title()} serializer won't sign "
                f"objects unless both "
                f"secret_key and secure_salt are set"
            )
        self.serializer = self.get_serializer(self.serializers[serializer_name])
        return await self.process(obj, **kwargs)  # type: ignore


dump = load = encode = decode = AcbEncoder()



import typing as t

from acb.config import ac
from pydantic import HttpUrl

from . import MonitoringBaseSettings


class MonitoringSettings(MonitoringBaseSettings):
    enabled: bool = ac.deployed or not ac.debug.production
    # dsn = "https://ea3f99402f144c2badf512c55d3d7bb7@o310698.ingest.sentry.io/1777286"
    dsn: HttpUrl
    sample_rate: int = 0.5

    def __init__(self, **values: t.Any) -> None:
        super().__init__(**values)
        self.sample_rate = self.sample_rate if ac.deployed else 1.0



import typing as t

from acb.config import ac
from acb.config import AppSettings
from acb.config import gen_password
from acb.config import load_adapter
from aiopath import AsyncPath
from pydantic import EmailStr
from pydantic import SecretStr
from pydantic import AnyUrl


class MailBaseSettings(AppSettings):
    api_key: SecretStr
    server: SecretStr
    password: SecretStr = gen_password(10)
    domain: EmailStr = f"mail@{ac.app.domain}"
    port: int = 587
    api_url: AnyUrl = None
    default_from: EmailStr = f"info@{ac.app.domain}"
    default_from_name: str = ac.app.title
    test_receiver: EmailStr = None
    tls: bool = True
    ssl: bool = False
    template_folder: t.Optional[AsyncPath] = None


mail = load_adapter("mail")



from contextlib import suppress
from time import sleep
from warnings import catch_warnings
from warnings import filterwarnings

from acb.config import ac
from google.api_core.exceptions import BadRequest
from google.api_core.exceptions import Conflict
from google.cloud.dns import Client as DnsClient
from validators import domain
from validators.utils import ValidationError
from . import DnsRecord
from . import DnsBaseSettings


# put into settings
with catch_warnings():
    filterwarnings("ignore", category=Warning)
    dns_client = DnsClient(project=ac.project)
dns_zone = dns_client.zone(ac.app_name, f"{ac.raw_domain}.")
splashstand_zone = dns_client.zone("sstand", "splashstand.com.")


class DnsSettings(DnsBaseSettings):
    ...


class Dns:
    @staticmethod
    def create_dns_zone() -> None:
        if not dns_zone.exists():
            print(f"Creating gdns zone '{ac.app_name}...")
            dns_zone.create()
            print(f"Zone '{dns_zone.name}' successfully created.")
        else:
            print(f"Zone for '{dns_zone.name}' exists.")

    @staticmethod
    async def list_dns_records(zone: DnsClient.zone = dns_zone):
        records = zone.list_resource_record_sets()
        records = [
            DnsRecord(
                name=record.name,
                type=record.record_type,
                ttl=record.ttl,
                rrdata=record.rrdatas,
            )
            for record in records
        ]
        # if ac.debug.dns or ac.debug.mail:
        #     await apformat(records)
        return records

    async def create_dns_records(
        self, records: list | DnsRecord, zone: DnsClient.zone = dns_zone
    ):
        if isinstance(records, DnsRecord):
            records = [records]
        changes = zone.changes()
        current_record_sets = []
        new_record_sets = []
        for record in records:
            if not record.name.endswith("."):
                record.name = f"{record.name}."
            record.rrdata = (
                [record.rrdata]
                if not isinstance(record.rrdata, list)
                else record.rrdata
            )
            for i, r in enumerate(record.rrdata):
                with suppress(ValidationError):
                    if isinstance(r, str) and domain(r) and not r.endswith("."):
                        r = f"{r}."
                        record.rrdata[i] = r
                    if record.type == "TXT":
                        record.rrdata[i] = f'"{r}"'
            current_record = [
                r
                for r in await self.list_dns_records()
                if r.name == record.name and r.type == record.type
            ]
            if len(current_record) == 1:
                # print(current_record[0])
                current_record = current_record[0]
                if current_record.__dict__ == record.__dict__:
                    continue
                current_record_set = zone.resource_record_set(
                    name=current_record.name,
                    record_type=current_record.type,
                    ttl=current_record.ttl,
                    rrdatas=current_record.rrdata,
                )
                current_record_sets.append(current_record_set)
                print(f"Deleting - {current_record}")
            record_set = zone.resource_record_set(
                name=record.name,
                record_type=record.type,
                ttl=record.ttl,
                rrdatas=record.rrdata,
            )
            new_record_sets.append(record_set)
            print(f"Creating - {record}")
        if current_record_sets:
            for record_set in current_record_sets:
                changes.delete_record_set(record_set)
            changes.create()  # API request
            print("Deleting record sets", end="")
            while changes.status != "done":
                print(".", end="")
                sleep(5)
                changes.reload()  # API request
            print()
        changes = zone.changes()
        if new_record_sets:
            for record_set in new_record_sets:
                changes.add_record_set(record_set)
            try:
                changes.create()  # API request
                print("Creating record sets", end="")
                while changes.status != "done":
                    print(".", end="")
                    sleep(5)
                    changes.reload()  # API request
                print()
            except (Conflict, BadRequest) as err:
                if changes.additions[0].name.split(".")[1] != "splashstand":
                    raise err
                print("SplashStand development domain detected. No changes made.")
        else:
            print("No DNS changes detected.")



from sqlalchemy.pool import NullPool
from . import SqlBase
from . import SqlBaseSettings
import typing as t


class SqlSettings(SqlBaseSettings):
    driver: str = "mysql+mysqldb"
    async_driver: str = "mysql+asyncmy"
    port: int = 3306
    pool_pre_ping: bool = True

    def model_post_init(self, __context: t.Any) -> None:
        super().model_post_init(self)
        self.poolclass = NullPool
        self.engine_kwargs = dict(
            poolclass=self.poolclass, pool_pre_ping=self.pool_pre_ping
        )


class Sql(SqlBase):
    ...


sql = Sql()



from . import MailBaseSettings


class MailSettings(MailBaseSettings):
    mx_servers = [
        "1 aspmx.l.google.com.",
        "5 alt1.aspmx.l.google.com.",
        "5 alt2.aspmx.l.google.com.",
        "10 alt3.aspmx.l.google.com.",
        "10 alt4.aspmx.l.google.com.",
    ]


class Mail:
    ...



import logging
import sys
from time import perf_counter

# from acb.config import ac
from loguru import logger


# def get_mod():
#     # frame = stack()[3][0]
#     # frame = logging.currentframe()
#     mod_logger = stack()[3][0]
#     mod = getmodule(mod_logger)
#     mod = Path(mod.__file__).parent
#     if mod.stem == ac.basedir.stem:
#         mod = Path(mod.__file__)
#     return mod
#
#
# def log_debug(s):
#     mod = get_mod()
#     debug_mod = dict(ac.debug.model_fields).get(mod.stem)
#     if debug_mod:
#         if ac.deployed:
#             return (
#             logger.patch(lambda record: record.update(
#             name=mod.__name__)).debug(s)
#             )
#         return colorizedStderrPrint(s)
#
#
# ic.configureOutput(prefix="    debug:  ", includeContext=True,
# outputFunction=log_debug)
# if ac.deployed:
#     ic.configureOutput(prefix="", includeContext=False, outputFunction=log_debug)


# async def apformat(obj, sort_dicts: bool = False) -> None:  # make purple
#     mod = get_mod()
#     debug_mod = dict(ac.debug.model_fields).get(mod.stem)
#     if not ac.deployed and not ac.debug.production and debug_mod:
#         await aprint(pformat(obj, sort_dicts=sort_dicts))


def timeit(func):
    def wrapped(*args, **kwargs):
        start = perf_counter()
        result = func(*args, **kwargs)
        end = perf_counter()
        logger.debug(f"Function '{func.__name__}' executed in {end - start} s")
        return result

    return wrapped


class InterceptHandler(logging.Handler):
    def __init__(self, logger_name: str) -> None:
        self.logger_name = logger_name
        super().__init__()

    def emit(self, record) -> None:
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        logger.patch(lambda record: record.update(name=self.logger_name)).opt(
            depth=depth, exception=record.exc_info
        ).log(level, record.getMessage())


logger_format = dict(
    time="<b><e>[</e> <w>{time:YYYY-MM-DD HH:mm:ss.SSS}</w> <e>]</e></b>",
    level=" <level>{level:>8}</level>",
    sep=" <b><w>in</w></b> ",
    name="{name:>20}",
    line="<b><e>[</e><w>{line:^5}</w><e>]</e></b>",
    message="  <level>{message}</level>",
)
# level_per_module = {
#     m: "DEBUG" if v is True else "INFO" for (m, v) in ac.debug.model_dump().items()
# }

log_format = "".join(logger_format.values())
configs = dict(
    # filter=level_per_module,
    format=log_format,
    enqueue=True,
    backtrace=True,
)
logger.remove()
# logging.getLogger("uvicorn").handlers.clear()
logger.add(sys.stderr, **configs)
logger.level("DEBUG", color="<cyan>")
_loggers = []
# _loggers = ["uvicorn.access", "uvicorn.error"]
# if ac.debug.sql:
#     _loggers.extend(
#         [
#             "sqlalchemy.engine",
#             "sqlalchemy.orm",
#             "sqlalchemy.pool",
#             "sqlalchemy.dialects",
#         ]
#     )
# if ac.debug.cache:
#     _loggers.extend(["httpx_caching"])
for _log in _loggers:
    _logger = logging.getLogger(_log)
    _logger.handlers.clear()
    _logger.handlers = [InterceptHandler(_logger.name)]

# if ac.debug.logger:
#     logger.debug("debug")
#     logger.info("info")
#     logger.warning("warning")
#     logger.error("error")
#     logger.critical("critical")
#     pprint(level_per_module)



from functools import cached_property
from functools import lru_cache

from acb.config import ac
from httpx import Response as HttpxResponse
from httpx_cache import AsyncClient
from httpx_cache.cache.redis import RedisCache
from pydantic import AnyHttpUrl
from . import RequestsBaseSettings


class RequestsSettings(RequestsBaseSettings):
    @cached_property
    def redis_connection(self):
        return AsyncClient(
            cache=RedisCache(
                redis_url=f"redis://{ac.cache.host}:{ac.cache.port}/{self.cache_db}"
            )
        )


class Requests:
    @staticmethod
    @lru_cache
    def connection():
        return RequestsSettings.redis_connection

    async def get(self, url: AnyHttpUrl) -> HttpxResponse:
        async with self.connection() as client:
            return await client.get(url)

    async def post(self, url: AnyHttpUrl, data: dict) -> HttpxResponse:
        async with self.connection() as client:
            return await client.post(url, data)

    async def put(self, url: AnyHttpUrl, data: dict) -> HttpxResponse:
        async with self.connection() as client:
            return await client.put(url, data)

    async def delete(self, url: AnyHttpUrl) -> HttpxResponse:
        async with self.connection() as client:
            return await client.delete(url)

    async def init(self) -> None:
        ...


requests = Requests()



import asyncio
import inspect
import sys
import typing as t
from abc import ABC
from abc import abstractmethod
from contextlib import suppress
from importlib import import_module
from pathlib import Path
from random import choice
from secrets import token_bytes
from secrets import token_urlsafe
from string import ascii_letters
from string import digits
from string import punctuation
from warnings import warn
from contextvars import ContextVar

import nest_asyncio
from acb.actions import dump
from acb.actions import load
from aiopath import AsyncPath
from async_lru import alru_cache
from icecream import ic
from inflection import camelize
from inflection import titleize
from inflection import underscore
from pydantic import AliasChoices
from pydantic import AliasPath
from pydantic import AnyUrl
from pydantic import BaseModel
from pydantic import SecretStr
from pydantic._internal._utils import deep_update
from pydantic.fields import FieldInfo
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict
from pydantic_settings.sources import _annotation_is_complex
from pydantic_settings.sources import SettingsError
from pydantic_settings.utils import path_type_label
from pydantic_settings.utils import path_type_labels

nest_asyncio.apply()

# deployed = True if basedir.name == "srv" else False
deployed = True if Path.cwd().name == "app" else False

project: str = ""
app_name: str = ""
app_secrets: set = set()
# enabled_adapters: dict = {}
enabled_adapters: ContextVar[dict] = ContextVar("enabled_adapters", default={})


class PackageRegistry:
    def __init__(self) -> None:
        self.packages: dict = {}

    def register(self) -> t.Any:
        pkg_path = AsyncPath(inspect.stack()[1][1]).parent
        pkg_name = pkg_path.stem
        self.packages[pkg_name] = pkg_path


package_registry = PackageRegistry()


def gen_password(size: int) -> str:
    chars = ascii_letters + digits + punctuation
    return "".join(choice(chars) for _ in range(size))


def load_adapter(adapter: str, settings: bool = False) -> t.Any:
    global enabled_adapters
    ic()
    ic(enabled_adapters)
    with suppress(KeyError):
        ic()
        ic(adapter)
        module = enabled_adapters.get()[adapter]
        ic(module)
        adapter_module = import_module(".".join(["acb", "adapters", adapter, module]))
        if settings:
            return getattr(adapter_module, adapter), getattr(
                adapter_module, f"{camelize(adapter)}Settings"
            )
        return getattr(adapter_module, adapter)
    warn(
        "AppConfig is not initialized. Please"
        " run `await ac.init()` before importing."
    )
    raise SystemExit(f"Adapter {adapter} not found.")


class PydanticBaseSettingsSource(ABC):
    def __init__(self, settings_cls: type["Settings"]) -> None:
        self.settings_cls = settings_cls
        self.config = settings_cls.model_config

    @abstractmethod
    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> tuple[t.Any, str, bool]:
        pass

    @staticmethod
    def field_is_complex(field: FieldInfo) -> bool:
        return _annotation_is_complex(field.annotation, field.metadata)

    async def prepare_field_value(
        self, field_name: str, field: FieldInfo, value: t.Any, value_is_complex: bool
    ) -> t.Any:
        if (self.field_is_complex(field) or value_is_complex) and value:
            return await load.json(value)
        return value

    @abstractmethod
    async def __call__(self) -> dict[str, t.Any]:
        pass


class InitSettingsSource(PydanticBaseSettingsSource):
    def __init__(
        self, settings_cls: type["Settings"], init_kwargs: dict[str, t.Any]
    ) -> None:
        self.init_kwargs = init_kwargs
        super().__init__(settings_cls)

    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> tuple[t.Any, str, bool]:
        return None, "", False

    async def __call__(self) -> dict[str, t.Any]:
        return self.init_kwargs

    def __repr__(self) -> str:
        return f"InitSettingsSource(init_kwargs={self.init_kwargs!r})"


class FileSecretsSource(PydanticBaseSettingsSource):
    def __init__(
        self,
        settings_cls: type["Settings"],
        secrets_dir: str | AsyncPath | None = None,
    ) -> None:
        super().__init__(settings_cls)
        self.secrets_dir = (
            secrets_dir if secrets_dir is not None else self.config.get("secrets_dir")
        )

    @staticmethod
    async def path_type_label(p: AsyncPath) -> str:
        if await p.exists():
            for method, name in path_type_labels.items():
                if getattr(p, method)():
                    return name
            return "unknown"

    @classmethod
    async def find_case_path(
        cls,
        dir_path: AsyncPath,
        file_name: str,
    ) -> AsyncPath | None:
        async for f in dir_path.iterdir():
            if f.name == file_name or f.name.lower() == file_name.lower():
                return f
        return None

    def _extract_field_info(
        self, field: FieldInfo, field_name: str
    ) -> list[tuple[str, str, bool]]:
        field_info: list[tuple[str, str, bool]] = []
        if isinstance(field.validation_alias, (AliasChoices, AliasPath)):
            v_alias: str | list[str | int] | list[
                list[str | int]
            ] | None = field.validation_alias.convert_to_aliases()
        else:
            v_alias = field.validation_alias
        if v_alias:
            if isinstance(v_alias, list):  # AliasChoices, AliasPath
                for alias in v_alias:
                    if isinstance(alias, str):  # AliasPath
                        field_info.append(
                            (
                                alias,
                                alias,
                                True if len(alias) > 1 else False,
                            )
                        )
                    elif isinstance(alias, list):  # AliasChoices
                        first_arg = t.cast(
                            str, alias[0]
                        )  # first item of an AliasChoices must be a str
                        field_info.append(
                            (
                                first_arg,
                                first_arg,
                                True if len(alias) > 1 else False,
                            )
                        )
            else:  # string validation alias
                field_info.append((v_alias, v_alias, False))
        else:
            field_info.append(
                (
                    field_name,
                    field_name,
                    False,
                )
            )
        return field_info

    async def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> tuple[t.Any, str, bool]:
        field_key: str = ""
        value_is_complex: bool = False
        for field_key, env_name, value_is_complex in self._extract_field_info(
            field, field_name
        ):
            path = await self.find_case_path(self.secrets_path, env_name)
            if not path:
                continue
            if await path.is_file():
                return (
                    SecretStr((await path.read_text()).strip()),
                    field_key,
                    value_is_complex,
                )
            else:
                warn(
                    f'attempted to load secret file "{path}" but found a '
                    f"{await self.path_type_label(path)} instead",
                    stacklevel=4,
                )
        return None, field_key, value_is_complex

    async def __call__(self) -> dict[str, t.Any]:
        global app_name, app_secrets
        data: dict[str, t.Any] = {}
        self.adapter_name: str = underscore(
            self.settings_cls.__name__.replace("Settings", "")
        )
        if self.adapter_name == "debug":
            return data
        if self.secrets_dir is None:
            return data
        self.secrets_path = await AsyncPath(self.secrets_dir).expanduser()
        if not await self.secrets_path.exists():
            warn(f'directory "{self.secrets_path}" does not exist')
            await self.secrets_path.mkdir(parents=True, exist_ok=True)
            # return data
        if not await self.secrets_path.is_dir():
            raise SettingsError(
                f"secrets_dir must reference a directory, not a "
                f"{path_type_label(self.secrets_path)}"
            )
        for field_name, field in self.settings_cls.model_fields.items():
            try:
                field_value, field_key, value_is_complex = await self.get_field_value(
                    field, "_".join([self.adapter_name, field_name])
                )
            except Exception as e:
                raise SettingsError(
                    f'error getting value for field "{field_name}" from '
                    f'source "{self.__class__.__name__}"'
                ) from e
            try:
                field_value = await self.prepare_field_value(
                    field_name, field, field_value, value_is_complex
                )
            except ValueError:
                raise SettingsError(
                    f'error parsing value for field "{field_name}" from'
                    f' source "{self.__class__.__name__}"'
                )
            if field_value is not None:
                app_secrets.add(field_key)
                data[field_name] = field_value
        return data

    def __repr__(self) -> str:
        return f"FileSecretsSource(secrets_dir={self.secrets_dir!r})"


class ManagerSecretsSource(FileSecretsSource):
    adapter_name: t.Optional[str] = None

    @alru_cache(maxsize=1)
    async def get_manager(self):
        global project, app_name, enabled_adapters
        ic()
        ic(enabled_adapters)
        manager = load_adapter("secrets")[0]
        await manager.init(project=project, app_name=app_name)
        return manager

    async def load_secrets(self) -> t.NoReturn:
        global project, app_name, app_secrets
        data: dict[str, t.Any] = {}
        model_secrets = {
            "_".join((self.adapter_name, n)): v
            for n, v in self.settings_cls.model_fields.items()
            if v.annotation is SecretStr
        }
        unfetched_secrets = {
            n: v for n, v in model_secrets.items() if n not in app_secrets
        }
        if unfetched_secrets:
            manager = await self.get_manager()
            manager_secrets = await manager.list(self.adapter_name)
            for field_key, field_value in unfetched_secrets.items():
                field_name = field_key.removeprefix(f"{self.adapter_name}_")
                stored_field_key = "_".join((app_name, field_key))
                secret_path = self.ac.secrets_path / field_key
                if not await secret_path.exists():
                    if field_key not in manager_secrets:
                        await manager.create(
                            stored_field_key, field_value.default.get_secret_value()
                        )
                    secret = await manager.get(stored_field_key)
                    await secret_path.write_text(secret)
                    data[field_name] = secret
                app_secrets.add(field_key)
        return data

    async def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> t.Tuple[t.Any, str, bool]:
        return field, field_name, False

    async def prepare_field_value(
        self, field_name: str, field: FieldInfo, value: t.Any, value_is_complex: bool
    ) -> str:
        return value

    async def __call__(self) -> dict[str, t.Any]:
        data: dict[str, t.Any] = {}
        self.adapter_name: str = underscore(
            self.settings_cls.__name__.replace("Settings", "")
        )
        if self.adapter_name == "debug":
            return data
        return await self.load_secrets()


class YamlSettingsSource(PydanticBaseSettingsSource):
    adapter_name: t.Optional[str] = None

    async def load_yml_settings(self) -> dict:
        global project, app_name
        if self.adapter_name == "secrets":
            return {}
        yml_path = AsyncPath.cwd() / "settings" / f"{self.adapter_name}.yml"
        if not await yml_path.exists() and not deployed:
            await yml_path.touch(exist_ok=True)
            dump_settings = {
                k: v.default
                for k, v in self.settings_cls.model_fields.items()
                if v.annotation is not SecretStr
            }
            await dump.yaml(dump_settings, yml_path)
        yml_settings = await load.yaml(yml_path)
        if self.adapter_name == "debug":
            for adptr in [
                adptr
                for adptr in enabled_adapters.get()
                if adptr not in yml_settings.keys()
            ]:
                yml_settings[adptr] = False
        if not deployed:
            await dump.yaml(yml_settings, yml_path)
        if self.adapter_name == "app":
            project = yml_settings["project"]
            app_name = yml_settings["name"]
        return yml_settings

    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> t.Tuple[t.Any, str, bool]:
        return field, field_name, False

    async def prepare_field_value(
        self, field_name: str, field: FieldInfo, value: t.Any, value_is_complex: bool
    ) -> t.Any:
        return value

    async def __call__(self) -> t.Dict[str, t.Any]:
        self.adapter_name: str = underscore(
            self.settings_cls.__name__.replace("Settings", "")
        )
        data: t.Dict[str, t.Any] = {}
        ic(Path.cwd())
        if Path.cwd().name != "acb":
            for field_name, field in (await self.load_yml_settings()).items():
                field_value, field_key, value_is_complex = self.get_field_value(
                    field, field_name
                )
                field_value = await self.prepare_field_value(
                    field_name, field, field_value, value_is_complex
                )
                if field_value is not None:
                    data[field_key] = field_value
            return data
        return data


class Settings(BaseModel):
    model_config: SettingsConfigDict = SettingsConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
        validate_default=True,
        secrets_dir=None,
        protected_namespaces=("model_", "settings_"),
    )

    def __init__(
        __pydantic_self__,
        _secrets_dir: str | Path | None = None,
        **values: t.Any,
    ) -> None:
        build_settings = __pydantic_self__._settings_build_values(
            values,
            _secrets_dir=_secrets_dir,
        )
        build_settings = asyncio.run(build_settings)
        super().__init__(**build_settings)

    async def _settings_build_values(
        self,
        init_kwargs: dict[str, t.Any],
        _secrets_dir: str | AsyncPath | None = None,
    ) -> dict[str, t.Any]:
        secrets_dir = (
            _secrets_dir
            if _secrets_dir is not None
            else self.model_config.get("secrets_dir")
        )
        init_settings = InitSettingsSource(self.__class__, init_kwargs=init_kwargs)
        file_secrets_settings = FileSecretsSource(
            self.__class__,
            secrets_dir=secrets_dir,
        )
        sources = self.settings_customise_sources(
            self.__class__,
            init_settings=init_settings,
            file_secrets_settings=file_secrets_settings,
        )
        if sources:
            return deep_update(*reversed([await source() for source in sources]))
        return {}

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: t.Type["Settings"],
        init_settings: PydanticBaseSettingsSource,
        file_secrets_settings: PydanticBaseSettingsSource,
    ) -> tuple[
        PydanticBaseSettingsSource,
        YamlSettingsSource,
        PydanticBaseSettingsSource,
        ManagerSecretsSource,
    ]:
        return (
            init_settings,
            YamlSettingsSource(settings_cls),
            file_secrets_settings,
            ManagerSecretsSource(settings_cls),
        )


class DebugSettings(Settings):
    production: bool = False
    main: bool = False
    secrets: bool = False
    logger: bool = False
    app: bool = False


class AppSettings(Settings):
    project: str = "myproject"
    name: str = "myapp"
    title: t.Optional[str] = None
    domain: t.Optional[AnyUrl] = None
    secret_key: SecretStr = SecretStr(token_urlsafe(32))
    secure_salt: SecretStr = SecretStr(str(token_bytes(32)))

    def model_post_init(self, __context: t.Any) -> None:
        self.title = self.title or titleize(self.name)


class AppConfig(BaseSettings, extra="allow"):
    pkgdir: AsyncPath = AsyncPath(__file__).parent
    basedir: AsyncPath = AsyncPath.cwd()
    deployed: bool = deployed
    tmp: t.Optional[AsyncPath] = None
    app_settings_path: t.Optional[AsyncPath] = None
    adapter_settings_path: t.Optional[AsyncPath] = None
    available_adapters: dict[str, t.Any] = {}
    adapter_categories: list = []
    # enabled_adapters: dict[str, t.Any] = {}
    secrets_path: t.Optional[AsyncPath] = None
    debug: t.Optional[t.Any] = DebugSettings()
    app: t.Optional[t.Any] = None

    def model_post_init(self, __context: t.Any) -> None:
        self.tmp = self.basedir / "tmp"
        self.app_settings_path = self.basedir / "settings"
        self.adapter_settings_path = self.app_settings_path / "adapters.yml"
        self.secrets_path = self.basedir / "tmp" / "secrets"
        self.deployed = True if self.basedir.name == "app" else deployed
        self.app = AppSettings(_secrets_dir=self.secrets_path)

    async def init(self, deployed: bool = False) -> "AppConfig":
        global enabled_adapters
        self.deployed = deployed
        for path in (
            self.app_settings_path,
            self.tmp,
            self.app_settings_path,
            self.secrets_path,
        ):
            await path.mkdir(exist_ok=True)
        mod_dir = self.basedir / "__pypackages__"
        sys.path.append(str(mod_dir))
        base_settings = {}
        if self.basedir.name != "acb":
            adapter_categories = [
                path.stem
                async for path in (self.pkgdir / "adapters").iterdir()
                if path.is_dir and not path.name.startswith("__")
            ]
            if not await self.adapter_settings_path.exists() and not self.deployed:
                await dump.yaml(
                    {cat: None for cat in adapter_categories},
                    self.adapter_settings_path,
                )
            available_adapters = await load.yaml(self.adapter_settings_path)
            enabled_adapters.set(
                {
                    c: a
                    for c, a in available_adapters.items()
                    if c in adapter_categories and a
                }
            )
            base_settings.update(
                dict(
                    available_adapters=available_adapters,
                    enabled_adapters=enabled_adapters,
                    adapter_categories=adapter_categories,
                )
            )
            super().__init__(**base_settings)
            del enabled_adapters.get()["secrets"]
        return self


# ac: ContextVar["AppConfig"] = ContextVar("ac", default=AppConfig())
ac = AppConfig()

# ac: AppConfig = dependency()



from acb.config import ac
from pydantic import BaseModel
from acb.config import Settings
from acb.config import load_adapter
import typing as t


class DnsRecord(BaseModel):
    name: str = ac.app.mail_domain
    type: str = "TXT"
    ttl: int = 300
    rrdata: t.Optional[str | list] = None


class DnsBaseSettings(Settings):
    ...


dns = load_adapter("dns")



from adlfs import AzureBlobFileSystem
from pydantic import SecretStr
from . import StorageBase
from . import StorageBaseSettings


class StorageSettings(StorageBaseSettings):
    connection_string: SecretStr


class Storage(StorageBase):
    client: AzureBlobFileSystem = AzureBlobFileSystem


storage = Storage()



import asyncio
import typing as t

import asyncio_atexit
from acb.actions import dump
from acb.actions import load
from acb.config import ac
from acb.logger import logger
from cashews.serialize import register_type
from cashews.wrapper import Cache as CashewsCache
from icecream import ic
from pydantic import RedisDsn
from . import CacheBaseSettings


class CacheSettings(CacheBaseSettings):
    def model_post_init(self, __context: t.Any) -> None:
        super().model_post_init(self)
        self.namespace = ac.app.name
        self._url: RedisDsn = (
            f"redis://{self.host.get_secret_value()}:{self.port}/{self.db}"
        )


class Cache(CashewsCache):
    @staticmethod
    async def encoder(value: t.Any, *args, **kwargs) -> bytes:
        return await dump.msgpack(
            value,
            secret_key=ac.app.secret_key.get_secret_value(),
            secure_salt=ac.app.secure_salt.get_secret_value(),
        )

    @staticmethod
    async def decoder(value: bytes, *args, **kwargs) -> t.Any:
        return await load.msgpack(
            value,
            secret_key=ac.app.secret_key.get_secret_value(),
            secure_salt=ac.app.secure_salt.get_secret_value(),
        )

    async def init(self, *args, **kwargs) -> t.NoReturn:
        await super().init(
            ac.cache._url,
            password=ac.cache.password.get_secret_value(),
            client_side=True,
            client_side_prefix=f"{ac.app.name}:",
        )
        register_type(t.Any, self.encoder, self.decoder)

        @asyncio_atexit.register
        async def close_cache_session() -> t.NoReturn:
            logger.debug("Closing cache session...")
            loop = asyncio.get_running_loop()
            ic(loop.is_running())
            await self.close()
            logger.debug("Cache session closed.")

        logger.debug("App cache initialized.")


cache = Cache()



import typing as t

from acb.config import ac
from acb.config import gen_password
from acb.config import Settings
from pydantic import AnyUrl
from pydantic import field_validator
from pydantic import RedisDsn
from pydantic import SecretStr


class CacheBaseSettings(Settings):
    namespace: t.Optional[str] = None
    db: int = 1
    host: SecretStr = SecretStr("127.0.0.1")
    password: SecretStr = SecretStr(gen_password(10))
    _url: t.Optional[RedisDsn | AnyUrl] = None
    default_timeout: int = 86400
    # template_timeout: int = 300 if ac.deployed else 1
    media_timeout: int = 15_768_000
    media_control: str = f"max-age={media_timeout} public"
    port: int = 6379
    health_check_interval: int = 15

    def model_post_init(self, __context: t.Any) -> None:
        self.host = SecretStr("127.0.0.1") if not ac.deployed else self.host
        self.password = SecretStr("") if not ac.deployed else self.password

    @field_validator("db")
    def db_less_than_three(cls, v) -> int:
        if v < 3 and v != 1:
            raise ValueError("must be greater than 2 (0-2 are reserved)")
        return 1



from acb.config import Settings
from pydantic import field_validator


class RequestsBaseSettings(Settings):
    cache_db: int = 2

    @field_validator("cache_db")
    def cache_db_less_than_three(cls, v) -> int:
        if v < 3 and v != 2:
            raise ValueError("must be greater than 2 (0-2 are reserved)")
        return 2


# requests = load_adapter("requests")



import asyncio
import typing as t

from acb.config import ac
from acb.config import Settings
from acb.logger import logger
from aiopath import AsyncPath
from google.cloud.exceptions import NotFound


# CORS policy for upload bucket - upload-cors.json
#
# [
#     {
#       "origin": ["*"],
#       "method": ["*"],
#       "responseHeader": ["*"],
#       "maxAgeSeconds": 600
#     }
# ]


class StorageBaseSettings(Settings):
    prefix: t.Optional[str] = None
    user_project: t.Optional[str] = None  # used for billing
    buckets: dict[str, str] = {}

    def model_post_init(self, __context: t.Any) -> None:
        self.prefix: str = ac.app.name
        self.user_project: str = ac.app.name  # used for billing


class StorageBucket:
    client: t.Any
    bucket: t.Optional[str]
    root: t.Optional[AsyncPath]
    prefix: str = ""
    cache_control: t.Optional[str] = None

    def __init__(self, client, bucket: str, prefix: str = None) -> None:
        self.client = client
        self.prefix = prefix or ac.app.name or self.prefix
        self.bucket = ac.storage.buckets[bucket]
        self.root = AsyncPath(f"{self.bucket}/{self.prefix}")

    def get_name(self, path: AsyncPath) -> str:
        return path.name

    def get_path(self, path: AsyncPath) -> str:
        return str(self.root / path)

    def get_url(self, path: AsyncPath):
        return self.client.url(self.get_path(path))

    async def get_date_created(self, path: AsyncPath) -> str:
        return (await self.stat(path))["timeCreated"]

    async def get_date_updated(self, path: AsyncPath) -> str:
        return (await self.stat(path))["updated"]

    async def get_size(self, path: AsyncPath) -> int:
        return (await self.stat(path))["size"]

    async def get_signed_url(self, path: AsyncPath, expires: int = 3600) -> str:
        return await self.client._sign(self.get_path(path), expires=expires)

    async def stat(self, path: AsyncPath) -> dict:
        return await self.client._info(self.get_path(path))

    async def list(self, dir_path: AsyncPath):
        return await self.client._ls(self.get_path(dir_path))

    async def exists(self, path: AsyncPath):
        return await self.client._exists(self.get_path(path))

    async def create_bucket(self, path: AsyncPath):
        return await self.client._mkdir(
            self.get_path(path),
            create_parents=True,
            enable_versioning=False,
            # acl="public-read",
        )

    async def read(self, path: AsyncPath):
        stor_path = self.get_path(path)
        logger.debug(f"Getting {stor_path}...")
        try:
            data = await self.client._cat_file(stor_path)
        except (NotFound, FileNotFoundError):
            raise FileNotFoundError  # for jinja loaders
        logger.debug(f"Got - {stor_path}")
        return data

    async def write(self, path: AsyncPath, data: t.Any) -> None:
        stor_path = self.get_path(path)
        logger.debug(f"Saving {stor_path}...")
        await self.client._pipe_file(stor_path, data)
        logger.debug(f"Saved - {stor_path}")

    async def delete(self, path: AsyncPath) -> None:
        stor_path = self.get_path(path)
        logger.debug(f"Deleting {stor_path}...")
        await self.client._rm_file(stor_path)
        logger.debug(f"Deleted - {stor_path}")


class StorageBase:
    client: t.Any

    # session: t.Any

    async def init(self) -> t.NoReturn:
        loop = asyncio.get_running_loop()
        self.client = self.client(asynchronous=True, loop=loop)
        for bucket in ac.storage.buckets:
            setattr(self, bucket, StorageBucket(self.client, bucket))
            logger.debug(f"{bucket.title()} storage bucket initialized.")
        logger.info("Storage initialized.")
        print(ac.storage.buckets)


# class BaseStorage:  # pragma: no cover
#     def get_name(self, name: str) -> str:
#         ...
#
#     def get_path(self, name: str) -> str:
#         ...
#
#     def get_size(self, name: str) -> int:
#         ...
#
#     def open(self, name: str) -> BinaryIO:
#         ...
#
#     def write(self, file: BinaryIO, name: str) -> str:
#         ...
#
#
# class StorageFile:
#     """
#     The file object returned by the storage.
#     """
#
#     def __init__(self, *, name: str, storage: BaseStorage) -> None:
#         self._name = name
#         self._storage = storage
#
#     @property
#     def name(self) -> str:
#         """File name including extension."""
#
#         return self._storage.get_name(self._name)
#
#     @property
#     def path(self) -> str:
#         """Complete file path."""
#
#         return self._storage.get_path(self._name)
#
#     @property
#     def size(self) -> int:
#         """File size in bytes."""
#
#         return self._storage.get_size(self._name)
#
#     def open(self) -> BinaryIO:
#         """
#         Open a file handle of the file.
#         """
#
#         return self._storage.open(self._name)
#
#     def write(self, file: BinaryIO) -> str:
#         """
#         Write input file which is opened in binary mode to destination.
#         """
#
#         return self._storage.write(file=file, name=self._name)
#
#     def __str__(self) -> str:
#         return self.path
#
#
# class StorageImage(StorageFile):
#     """
#     Inherits features of `StorageFile` and adds image specific properties.
#     """
#
#     def __init__(
#         self, *, name: str, storage: BaseStorage, height: int, width: int
#     ) -> None:
#         super().__init__(name=name, storage=storage)
#         self._width = width
#         self._height = height
#
#     @property
#     def height(self) -> int:
#         """
#         Image height in pixels.
#         """
#
#         return self._height
#
#     @property
#     def width(self) -> int:
#         """
#         Image width in pixels.
#         """
#
#         return self._width



import asyncio
import sys
import typing as t
from pprint import pformat
from pprint import pprint
from re import search

from acb.actions import load
from acb.adapters.dns import dns
from acb.adapters.dns import DnsRecord
from acb.adapters.requests import requests
from acb.config import ac
from aiopath import AsyncPath
from httpx import Response as HttpxResponse
from loguru import logger
from pydantic import AnyUrl
from pydantic import EmailStr
from pydantic import SecretStr
from . import MailBaseSettings


class MailSettings(MailBaseSettings):
    password: SecretStr
    api_key: SecretStr
    server: SecretStr = "smtp.mailgun.com"
    api_url = "https://api.mailgun.net/v3/domains"
    default_from: EmailStr = f"info@{ac.app.domain}"
    default_from_name = ac.app.title
    test_receiver: EmailStr = None
    tls = True
    ssl = False
    template_folder: t.Optional[AsyncPath]


class Mail:
    async def get_response(
        self,
        req_type: str,
        domain: AnyUrl = None,
        data: dict = None,
        params: dict = None,
    ) -> dict:
        caller: str = sys._getframe().f_back.f_code.co_name
        match caller:
            case search(caller, "domain"):
                caller = "domain"
            case search(caller, "route"):
                caller = "route"
        url = "/".join([ac.mailgun.api_url, caller, domain])
        data = dict(auth=("api", ac.mailgun.api_key), params=params, data=data)
        resp = None
        match req_type:
            case "get":
                resp = await requests.get(url)
            case "put":
                url = "".join([url, "/connection"])
                resp = await requests.put(url, data=data)
            case "post":
                url = "".join([url, "/connection"])
                resp = await requests.post(url, data=data)
            case "delete":
                resp = await requests.delete(url)
        if ac.debug.mail:
            print(resp)
        return await load.json(resp.json())

    async def list_domains(self) -> list:
        resp = await self.get_response("get", params={"skip": 0, "limit": 1000})
        domains = [d["name"] for d in resp["items"]]
        return domains

    async def get_domain(self, domain: AnyUrl) -> dict:
        return await self.get_response("get", domain=domain)

    async def create_domain(self, domain: AnyUrl) -> dict:
        resp = await self.get_response(
            "post",
            data={
                "name": domain,
                "smtp_password": ac.mail.password,
                "web_scheme": "https",
            },
        )
        if ac.debug.mail:
            logger.debug(resp)
        resp = await self.get_response(
            "put", data={"require_tls": True, "skip_verification": True}
        )
        return resp

    async def delete_domain(self, domain: AnyUrl) -> dict:
        return await self.get_response("delete", domain=domain)

    async def create_domain_credentials(self, domain: AnyUrl) -> dict:
        return await self.get_response(
            "post",
            domain=domain,
            data={"login": f"postmaster@{domain}", "password": ac.mail.password},
        )

    async def update_domain_credentials(self, domain: AnyUrl) -> dict:
        # this is prob not necessary
        return await self.get_response(
            "put", domain=domain, data={"password": ac.mail.mailgun.password}
        )

    async def get_dns_records(self, domain: AnyUrl) -> list:
        records = await self.get_domain(domain)
        mx_records = records["receiving_dns_records"]
        sending_records = records["sending_dns_records"]
        records = []
        rrdata = []
        for r in mx_records:
            mx_host = f"{r['priority']} {r['value']}."
            rrdata.append(mx_host)
        record = DnsRecord(name=domain, type="MX", rrdata=rrdata)
        records.append(record)
        for r in sending_records:
            record = DnsRecord(name=r["name"], type=r["record_type"], rrdata=r["value"])
            records.append(record)
        return records

    async def create_dns_records(self) -> None:
        # if not ac.mail.mailgun.domain in list_domains(self):
        await self.create_domain(ac.mail.mailgun.domain)
        await self.create_domain_credentials(ac.mail.mailgun.domain)
        records = await self.get_dns_records(ac.mail.mailgun.domain)
        if ac.mail.mailgun.gmail.enabled:
            await self.delete_domain(ac.mail.mailgun.domain)
            rrdata = ac.mail.gmail.mx_servers
            record = DnsRecord(name=ac.mail.mailgun.domain, type="MX", rrdata=rrdata)
            records.append(record)
        else:
            await self.delete_domain(ac.mail.mailgun.domain)
        if ac.debug.mail:
            pprint(records)
            await dns.create_dns_records(records)

    async def list_routes(self) -> list:
        resp = await self.get_response("get", params={"skip": 0, "limit": 1000})
        domain_routes = [
            r for r in resp["items"] if ac.mail.mailgun.domain in r["expression"]
        ]
        if ac.debug.mail:
            logger.debug(pformat(resp["items"]))
            logger.debug(len(resp["items"]))
            logger.debug(pformat(domain_routes))
            logger.debug(len(domain_routes))
        return domain_routes

    async def delete_route(self, route: dict) -> HttpxResponse:
        resp = await requests.delete("delete", domain={route["id"]})
        logger.info(f"Deleted route for {route['description']}")
        return resp

    @staticmethod
    def get_name(address: str) -> str:
        name = search("'(.+)@.+", address)
        return name.group(1) if name else ""

    async def delete_routes(self, delete_all: bool = False) -> None:
        forwards = ac.mail.forwards.keys()
        routes = await self.list_routes()
        deletes = []
        deletes.extend(
            [r for r in routes if self.get_name(r["expression"]) not in forwards]
        )
        if ac.debug.mail:
            pprint(deletes)
        for f in forwards:
            fs = [r for r in routes if self.get_name(r["expression"]) == f]
            deletes.extend(fs[1:])
        if delete_all or ac.mail.mailgun.gmail.enabled:
            deletes = [r for r in routes if len(self.get_name(r["expression"]))]
        if ac.debug.mail:
            pprint(deletes)
            print(len(deletes))
        else:
            for d in deletes:
                await self.delete_route(d)

    async def create_route(self, domain_address, forwarding_addresses) -> dict | None:
        domain_address = f"{domain_address}@{ac.mail.mailgun.domain}"
        if not isinstance(forwarding_addresses, list):
            forwarding_addresses = [forwarding_addresses]
        actions = ["stop(self)"]
        for addr in forwarding_addresses:
            actions.insert(0, f"forward('{addr}')")
        if ac.debug.mail:
            print(actions)
        route = {
            "priority": 0,
            "description": domain_address,
            "expression": f"match_recipient('{domain_address}')",
            "action": actions,
        }
        routes = await self.list_routes()
        await asyncio.gather(*routes)
        for r in routes:
            if (r["expression"] == route["expression"]) and (
                r["actions"] == route["action"]
            ):
                print(
                    f"Route for {domain_address}  ==> "
                    f"{', '.join(forwarding_addresses)} exists"
                )
                return
            elif r["expression"] == route["expression"]:
                await self.delete_route(r)
                break
        resp = await self.get_response("post", data=route)
        logger.info(
            f"Created route for {domain_address}  ==> "
            f" {', '.join(forwarding_addresses)}"
        )
        return resp

    async def create_routes(self, ordered: bool = True) -> None:
        if ac.mail.mailgun.gmail.enabled:
            logger.info("Using gmail mx servers. No routes created.")
            return
        else:
            forwards = await load.yaml(AsyncPath("mail.yml"), ordered=ordered)
            async for k, v in forwards.items():
                await self.create_route(k, v)

    async def setup_mail(self) -> None:
        await self.create_dns_records()
        await self.delete_routes()
        await self.create_routes()



from pydantic import SecretStr
from s3fs import S3FileSystem
from . import StorageBase
from . import StorageBaseSettings


class StorageSettings(StorageBaseSettings):
    access_key_id: SecretStr
    secret_access_key: SecretStr


class Storage(StorageBase):
    client: S3FileSystem = S3FileSystem


storage = Storage()



import asyncio

from acb.config import ac
from acb.config import load_adapter
from acb.config import enabled_adapters

__all__ = []


async def main() -> None:
    await ac.init()
    for adapter in enabled_adapters.get():
        __all__.append(adapter)
        globals()[adapter], module_settings = load_adapter(adapter, settings=True)
        setattr(ac, adapter, module_settings(_secrets_dir=ac.secrets_path))
        await globals()[adapter].init()


loop = asyncio.new_event_loop() or asyncio.get_running_loop()
loop.run_until_complete(main())



import typing as t
from calendar import isleap
from calendar import monthrange

# from concurrent.futures import as_completed
# from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager

# from contextlib import suppress
from datetime import date
from datetime import datetime
from datetime import timedelta
from functools import cached_property
from functools import lru_cache
from itertools import chain

# from pathlib import Path
from re import search

import arrow
from acb.config import ac
from acb.config import gen_password
from acb.config import Settings

# from acb.logger import apformat
from acb.logger import logger
from aioconsole import ainput
from aioconsole import aprint
from aiopath import AsyncPath
from pydantic import BaseModel
from pydantic import SecretStr

# from pydantic import create_model
from sqlalchemy import inspect
from sqlalchemy import text
from sqlalchemy.engine import URL
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.ext.asyncio import create_async_engine

# from sqlalchemy.exc import IntegrityError
# from sqlalchemy.exc import InvalidRequestError
# from sqlalchemy.ext.serializer import dumps as sdumps
# from sqlalchemy.ext.serializer import loads as sloads
from sqlalchemy_utils import create_database
from sqlalchemy_utils import database_exists
from sqlalchemy_utils import drop_database


# import arrow
# from acb.actions import dump
# from acb.actions import load
# from inflection import underscore
# from sqlalchemy.ext.hybrid import hybrid_property
# from sqlalchemy.orm import declared_attr
# from sqlmodel import Field
# from ulid import ULID
# from sqlmodel.ext.asyncio.session import AsyncSession
# from sqlmodel import select
# from sqlmodel import SQLModel


class SqlBaseSettings(Settings):
    driver: str
    async_driver: str
    port: int
    pool_pre_ping: bool = False
    poolclass: t.Any = None
    host: SecretStr = SecretStr("127.0.0.1")
    user: SecretStr = SecretStr("root")
    password: SecretStr = SecretStr(gen_password(10))
    url: t.Optional[URL] = None
    async_url: t.Optional[URL] = None

    @cached_property
    def async_engine(self) -> AsyncEngine:
        return create_async_engine(self.async_url, **self.engine_kwargs)

    # @cached_property
    # def async_session(self) -> AsyncSession:
    #     return AsyncSession(self.async_engine, expire_on_commit=False)

    def model_post_init(self, __context: t.Any) -> t.NoReturn:
        url_kwargs = dict(
            drivername=self.driver,
            username=self.user.get_secret_value(),
            password=self.password.get_secret_value(),
            host="127.0.0.1" if not ac.deployed else self.host.get_secret_value(),
            port=self.port,
            database=ac.app.name,
        )
        self.url = URL.create(**url_kwargs)
        async_url_kwargs = dict(drivername=self.async_driver)
        self.async_url = URL.create(**(url_kwargs | async_url_kwargs))


class SqlBase:
    async def create(self, demo: bool = False) -> t.NoReturn:
        exists = database_exists(ac.sql.url)
        if exists:
            logger.debug("Database exists.")

        if ac.debug.sql and not (ac.deployed or ac.debug.production) and exists:
            msg = (
                "\n\nRESETTING THE DATABASE WILL CAUSE ALL OF YOUR"
                " CURRENT DATA TO BE LOST!\n"
            )
            if demo:
                msg = (
                    "\nBy running this module,\n\nYOUR DATABASE WILL BE DELETED AND "
                    "REPLACED WITH DEMO DATA.\nALL OF YOUR CURRENT DATA WILL BE LOST!\n"
                )
            await aprint(msg)
            delete_db = await ainput("Would you like to reset the database? (Y/N) ")
            if delete_db:
                drop_database(ac.sql.url)
                logger.warning("Database dropped.")
                exists = database_exists(ac.sql.url)

        if not exists:
            create_database(ac.sql.url)
            logger.info("Database created.")

    # @lru_cache
    # def get_async_session(self) -> AsyncSession:
    #     return ac.db.async_session

    # @asynccontextmanager
    # async def session(self) -> t.AsyncGenerator:
    #     async with self.get_async_session() as sess:
    #         yield sess

    @lru_cache
    def get_async_engine(self) -> AsyncEngine:
        return ac.sql.async_engine

    @asynccontextmanager
    async def engine(self) -> t.AsyncGenerator:
        async with self.get_async_engine() as conn:
            yield conn

    @staticmethod
    def get_table_names(conn) -> list[str]:
        inspector = inspect(conn)
        return inspector.get_table_names()

    async def init(self, demo: bool = False) -> t.NoReturn:
        # print(debug.database)
        # print(type(debug.database))
        await self.create(demo)
        if ac.debug.sql:
            sql = text("DROP TABLE IF EXISTS alembic_version")
            self.get_async_engine.execute(sql)
        logger.info("Creating database tables...")
        # self.get_async_engine().run_sync(SQLModel.metadata.create_all)
        # if ac.debug.sql:
        #     table_names = self.get_async_engine().run_sync(self.get_table_names)
        #     await apformat(table_names)
        logger.info("Database initialized.")


# class AppBaseModel(SQLModel):
#     __table_args__ = {"extend_existing": True}
#     __mapper_args__ = {"always_refresh": True}
#     id: t.Optional[ULID] = Field(default_factory=ULID, primary_key=True)
#
#     @property
#     def date_created(self):
#         return arrow.get(ULID(self.id).timestamp)
#
#     @declared_attr
#     def __tablename__(cls):
#         return underscore(cls.__name__)
#
#     class Config:
#         arbitrary_types_allowed = True
#         extra = "allow"
#         json_loads = load.json
#         json_dumps = dump.json
#
#     async def save(self) -> None:
#         async with db.async_session() as session:
#             session.add(self)
#             await session.commit()


sure_delete = False


class BackupDbUtils(BaseModel):
    @staticmethod
    def get_timestamp(name: str) -> object:
        pattern = r"(.*)(-)(?P<timestamp>[\d]{10})(-)(.*)"
        match = search(pattern, name)
        return match.group("timestamp") if match else False

    # def get_files(self):
    #     blobs = stor.db.list()
    #     for blob in blobs:
    #         if self.get_timestamp(blob.name):
    #             yield blob.name

    def get_timestamps(self) -> list[object]:
        if not self.files:
            self.files = tuple(self.get_files())
        different_timestamps = []
        for name in self.files:
            timestamp = self.get_timestamp(name)
            if timestamp and timestamp not in different_timestamps:
                different_timestamps.append(timestamp)
        return different_timestamps

    def by_timestamp(self, timestamp: object) -> t.Generator:
        if not self.files:
            self.files = tuple(self.get_files())
        for name in self.files:
            if timestamp == self.get_timestamp(name):
                yield name

    def valid(self, timestamp: str) -> bool:
        if timestamp in self.get_timestamps():
            return True
        # print('==> Invalid id. Use "history" to list existing downloads')
        return False

    def get_path(self, class_name: str, timestamp: int = None) -> AsyncPath:
        timestamp = timestamp or arrow.utcnow().int_timestamp
        self.backup_path = AsyncPath(f"{ac.app.name}-{timestamp}-{class_name}.sqla")
        return self.backup_path


class BackupDbDates(BaseModel, arbitrary_types_allowed=True):
    today: arrow.Arrow = arrow.utcnow()
    white_list: list = []
    black_list: list = []
    dates: list = []

    def __init__(self, dates=None, **data: t.Any) -> None:
        super().__init__(**data)
        self.dates = sorted(dates, reverse=True) if dates else []
        self.run()  # feed self.white_list & self.black_list

    def get_last_month_length(self) -> datetime.month:
        return monthrange(self.today.year, self.today.shift(months=-1).month)[1]

    def get_last_year_length(self):
        first_day = date(self.today.year, 1, 1)  # current year
        last_day = first_day - timedelta(days=1)  # last year
        return 366 if isleap(last_day.year) else 365

    def filter_dates(self, dates, period) -> t.Coroutine:
        reference = self.today.int_timestamp
        method_mapping = {
            "week": lambda obj: getattr(obj, "isocalendar")()[1],
            "month": lambda obj: getattr(obj, "month"),
            "year": lambda obj: getattr(obj, "year"),
        }
        for dt in dates:
            comp_date = datetime.fromtimestamp(int(dt))
            comparison = method_mapping.get(period)(comp_date)
            ref_date = datetime.fromtimestamp(reference)
            if comparison != method_mapping.get(period)(ref_date):
                reference = dt
                yield dt

    def run(self) -> t.NoReturn:
        last_w = self.today.shift(days=-7).int_timestamp
        last_m = self.today.shift(days=-(self.get_last_month_length())).int_timestamp
        last_y = self.today.shift(days=-(self.get_last_year_length())).int_timestamp
        backups_week = []
        backups_month = []
        backups_year = []
        backups_older = []
        for timestamp in self.dates:
            if int(timestamp) >= last_w:
                backups_week.append(timestamp)
            elif int(timestamp) >= last_m:
                backups_month.append(timestamp)
            elif int(timestamp) >= last_y:
                backups_year.append(timestamp)
            else:
                backups_older.append(timestamp)
        self.white_list.extend(
            chain(
                backups_week,
                self.filter_dates(backups_month, "week"),
                self.filter_dates(backups_year, "month"),
                self.filter_dates(backups_older, "year"),
            )
        )
        diff_as_list = [d for d in self.dates if d not in self.white_list]
        self.black_list.extend(sorted(diff_as_list, reverse=True))


# class BackupDb(BackupDbDates, BackupDbUtils):
#     do_not_backup: list = []
#     models: list = []
#
#     def show(self):
#         return [
#             m
#             for m in SQLModel.metadata.schema
#             if isinstance(m, type) and issubclass(m, AppBaseModel)
#         ]
#
#     def get_mapped_classes(self) -> list[str]:
#         self.add_subclasses(AppBaseModel)
#         return self.models
#
#     def add_subclasses(self, model) -> None:
#         if model.__subclasses__():
#             for submodel in model.__subclasses__():
#                 self.add_subclasses(submodel)
#         else:
#             self.models.append(model)
#
#     async def get_data(self) -> dict[str, str]:
#         data = {}
#         async with db.session() as session:
#             for model in self.get_mapped_classes():
#                 query = select(model)
#                 results = session.exec(query)
#                 data[model.__name__] = sdumps(results)
#         return data
#
#     def parse_data(self, contents) -> bytes:
#         with suppress(AttributeError):
#             contents = sloads(
#                 contents,
#                 metadata=SQLModel.metadata,
#                 engine=db.engine,
#                 scoped_session=db.session(),
#             )
#         return contents
#
#     async def backup(self, class_name: str, data, now) -> None:
#         path = self.get_path(class_name, now)
#         await stor.db.save(data, path)
#
#     async def save(self) -> None:
#         data = self.get_data()
#         now = arrow.utcnow().int_timestamp
#
#         for class_name in data.keys():
#             path = self.get_path(class_name, now)
#             logger.debug(f"Backing up - {path.name}")
#             await self.backup(class_name, data[class_name], now)
#
#     async def get_backups(self) -> list[str]:
#         timestamps = self.get_timestamps()
#         backups = {}
#         for timestamp in timestamps:
#             adate = arrow.Arrow.fromtimestamp(timestamp)
#             date_formatted = adate.humanize()
#             backups[timestamp] = []
#             fps = [AsyncPath(b) for b in stor.db.list() if timestamp in b.name]
#             for fp in fps:
#                 if await stor.db.exists(fp):
#                     backups[timestamp].append(fp.name)
#             adate_formatted = adate.format("MM-DD-YYYY HH:mm:ss ZZ")
#             logger.debug(f"{date_formatted} ==> {adate_formatted}:")
#         return sorted(backups, reverse=True)
#
#     def get_last_backup(self) -> str | bool:
#         timestamps = self.get_timestamps()
#         if not len(timestamps):
#             return False
#         timestamp = sorted(timestamps, reverse=True)[0]
#         adate = arrow.Arrow.fromtimestamp(timestamp)
#         logger.info(f"Last backup: {adate.format('MM-DD-YYYY HH:mm:ss ZZ')}")
#         return timestamp
#
#     def restore_rows(self, path: Path) -> None:
#         stor.db.get(path)
#         # try:
#         #     model = search("-(\D+).sqla", path.name).group(1)
#         #     if contents:
#         #         contents = sub(
#         #             b"_sp.{31,32}\.",
#         #             bytes(f"{plugin_source.spaceid}.", encoding="utf8"),
#         #             contents,
#         #         )
#         #         is_pickled = load.pickle(contents)
#         #         if not isinstance(is_pickled, list):
#         #             contents = is_pickled
#         #         loaded = self.parse_data(contents)
#         #         if "app" in path.stem:
#         #             loaded = [loaded[0]]
#         #         if loaded != []:
#         #             for m in self.show():
#         #                 m_name = m.__name__
#         #                 if m_name == model:
#         #                     self.async_session.query(m).delete()
#         #             self.async_session.commit()
#         #             for row in loaded:
#         #                 try:
#         #                     with suppress(UnmappedInstanceError):
#         #                         self.async_session.merge(row)
#         #                     self.async_session.commit()
#         #                 except (IntegrityError, InvalidRequestError) as err:
#         #                     self.async_session.rollback()
#         #                     fails.append(row)
#         #             status = "partially" if len(fails) else "totally"
#         #             status = f"\t==> {path.name} {status} " \
#         #                      f"restored{':' if len(fails) else '.'}"
#         #         else:
#         #             status = f"\t!=> No or bad data for {model}."
#         #     else:
#         #         status = f"\t!=> File {path.name} does not exist."
#         # except Exception as err:
#         #     status = err
#         # return status, fails
#
#     def restore_backup(self, timestamp: str) -> None:
#         paths = [Path(b.name) for b in stor.db.list() if timestamp in b.name]
#         when = arrow.Arrow.fromtimestamp(timestamp).humanize()
#         logger.info(f"Restoring backup from {when}.....")
#         with ThreadPoolExecutor(max_workers=5) as executor:
#             all_tasks = []
#             for path in paths:
#                 logger.debug(f"\tRestoring:  {path.name}")
#                 all_tasks.append(executor.submit(self.restore_rows, path))
#
#             for future in as_completed(all_tasks):
#                 status, fails = future.result()
#                 for f in fails:
#                     logger.error(f"\t\tRestore of {f} failed!")
#
#     def delete_backups(self, delete_list: list[str]) -> None:
#         delete_me = stor.db.delete(delete_list)
#         if delete_me:
#             for name in delete_list:
#                 logger.debug(f"Deleted {name}")
#
#     def clean(self) -> str | bool:
#         """
#         Remove a series of backup files based on the following rules:
#         * Keeps all the backups from the last 7 days
#         * Keeps the most recent backup from each week of the last month
#         * Keeps the most recent backup from each month of the last year
#         * Keeps the most recent backup from each year of the remaining years
#         """
#
#         # get black and white list
#         cleaning = BackupDbDates(self.get_timestamps())
#         white_list = cleaning.white_list
#         black_list = cleaning.black_list
#         if not black_list:
#             logger.debug("==> No backup to be deleted.")
#             return True
#
#         logger.debug(f"==> {len(white_list)} backups will be kept:")
#         for timestamp in white_list:
#             date_formatted = arrow.Arrow.fromtimestamp(timestamp).humanize()
#             logger.debug(f"ID: {timestamp} (from {date_formatted})")
#             if ac.debug.database:
#                 for f in self.by_timestamp(timestamp):
#                     logger.debug(f)
#
#         delete_list = []
#         logger.debug(f"==> {len(black_list)} backups will be deleted:")
#         for timestamp in black_list:
#             date_formatted = arrow.Arrow.fromtimestamp(timestamp).humanize()
#             logger.debug(f"ID: {timestamp} (from {date_formatted})")
#             for f in self.by_timestamp(timestamp):
#                 if ac.debug.database:
#                     logger.debug(f)
#                 delete_list.append(f)
#
#         self.delete_backups(delete_list)
#         return "cleandb"
#
#     async def run(self) -> None:
#         if not ac.deployed:
#             last_backup = self.get_last_backup()
#             if ac.debug.database and sure_delete:
#                 blobs = [b.name for b in stor.db.list()]
#                 for b in blobs:
#                     stor.db.delete(b)
#                 # clear_resized_images()
#                 last_backup = None
#             if not last_backup:
#                 await self.save()
#             elif last_backup and ac.debug.database and not sure_delete:
#                 logger.info(f"Restoring last backup - {last_backup}")
#                 self.restore_backup(last_backup)
#             self.clean()
#             logger.info("Backups complete.")
#
#
# backup_db = BackupDb()



import typing as t
from contextlib import suppress
from secrets import compare_digest
from typing import Optional
from warnings import catch_warnings
from warnings import filterwarnings

from acb.config import ac
from acb.logger import logger
from google.api_core.exceptions import AlreadyExists
from google.auth import default
from google.auth.transport.requests import AuthorizedSession
from google.auth.transport.requests import Request
from google.cloud.secretmanager_v1 import AccessSecretVersionRequest
from google.cloud.secretmanager_v1 import AddSecretVersionRequest
from google.cloud.secretmanager_v1 import CreateSecretRequest
from google.cloud.secretmanager_v1 import DeleteSecretRequest
from google.cloud.secretmanager_v1 import ListSecretsRequest
from google.cloud.secretmanager_v1 import SecretManagerServiceAsyncClient
from google.oauth2.credentials import Credentials
from . import SecretsBaseSettings


class SecretsSettings(SecretsBaseSettings):
    ...


class Secrets:
    parent: str
    client: Optional[SecretManagerServiceAsyncClient]
    authed_session: Optional[AuthorizedSession]
    creds: Optional[Credentials]

    # @staticmethod
    def extract_secret_name(self, secret_path: str) -> str:
        return secret_path.split("/")[-1].removeprefix(self.prefix)

    async def get_access_token(self) -> str:
        self.creds.refresh(Request())
        logger.debug(f"Secrets access token:\n{self.creds.token}")
        return self.creds.token

    async def verify_access_token(self, token: str) -> bool:
        verified = compare_digest(self.creds.token, token)
        logger.debug(f"Secrets access token verified - {verified}")
        return verified

    async def list(self, adapter: str) -> list:
        request = ListSecretsRequest(
            parent=self.parent, filter=f"{self.prefix}{adapter}_"
        )
        client_secrets = await self.client.list_secrets(request=request)
        client_secrets = [
            self.extract_secret_name(secret.name) async for secret in client_secrets
        ]
        return client_secrets

    async def get(self, name: str) -> str:
        path = f"projects/{self.project}/secrets/{name}/versions/latest"
        request = AccessSecretVersionRequest(name=path)
        version = await self.client.access_secret_version(request=request)
        payload = version.payload.data.decode()
        logger.info(f"Fetched secret - {name}")
        return payload

    async def create(self, name: str, value: str) -> t.NoReturn:
        with suppress(AlreadyExists):
            request = CreateSecretRequest(
                parent=self.parent,
                secret_id=name,
                secret={"replication": {"automatic": {}}},
            )
            version = await self.client.create_secret(request)
            request = AddSecretVersionRequest(
                parent=version.name,
                payload={"data": f"{value}".encode()},
            )
            await self.client.add_secret_version(request)
            if not ac.deployed:
                logger.debug(f"Created secret - {name}")

    async def update(self, name: str, value: str) -> t.NoReturn:
        secret = self.client.secret_path(self.project, name)
        request = AddSecretVersionRequest(
            parent=secret,
            payload={"data": f"{value}".encode()},
        )
        await self.client.add_secret_version(request)
        if not ac.deployed:
            logger.debug(f"Updated secret - {name}")

    async def delete(self, name: str) -> t.NoReturn:
        secret = self.client.secret_path(self.project, name)
        request = DeleteSecretRequest(name=secret)
        await self.client.delete_secret(request=request)
        if not ac.deployed:
            logger.debug(f"Deleted secret - {secret}")

    async def init(self, project: str, app_name: str) -> t.NoReturn:
        self.project = project
        self.parent = f"projects/{project}"
        self.app_name = app_name
        self.prefix = f"{app_name}_"
        with catch_warnings():
            filterwarnings("ignore", category=Warning)
            creds, _ = default()
        self.creds = creds
        self.client = SecretManagerServiceAsyncClient(credentials=self.creds)
        self.authed_session = AuthorizedSession(self.creds)


secrets = Secrets()



from gcsfs.core import GCSFileSystem
from . import StorageBase
from . import StorageBaseSettings


class StorageSettings(StorageBaseSettings):
    ...


class Storage(StorageBase):
    client: GCSFileSystem = GCSFileSystem


storage = Storage()
